import os

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarOverride(Rule):
    def __init__(self):
        super().__init__(id='oelint.var.override',
                         severity='error',
                         message='<FOO>')

    def check(self, _file, stash):
        res = []
        __varnames = [x.VarName for x in stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)]
        for v in __varnames:
            if v == 'inherit' or not v:
                # This will be done by another rule
                continue  # pragma: no cover
            items = stash.GetItemsFor(
                filename=_file, classifier=Variable.CLASSIFIER, attribute=Variable.ATTR_VAR, attributeValue=v)
            items = sorted(items, key=lambda x: x.Line, reverse=False)
            for sub in [x.SubItem for x in items]:
                # Get all entries but not the only that do immediate expansion,
                # as these will be handled during parse time
                # and apply to different rules
                _items = [x for x in items if x.SubItem == sub and not x.IsAppend(
                ) and x.VarOp not in [' := ', ' ?= ', ' ??= '] and not x.Flag]
                if len(_items) > 1:
                    _files = {os.path.basename(x.Origin) for x in _items}
                    res += self.finding(_items[0].Origin, _items[0].InFileLine,
                                        'Variable \'{a}\' is set by \'{b}\''.format(
                                        a=v, b=','.join(_files)))
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarFilesOverride(Rule):
    def __init__(self):
        super().__init__(id='oelint.var.filesoverride',
                         severity='warning',
                         message='\'{a}\' should not be overriden')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if i.VarName not in ['FILES']:
                continue
            if not i.AppendOperation():
                res += self.finding(i.Origin, i.InFileLine,
                                    override_msg=self.Msg.format(a=i.VarName))
        return res



import ast

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Function


class TaskNoPythonPrefix(Rule):
    def __init__(self):
        super().__init__(id='oelint.task.nopythonprefix',
                         severity='warning',
                         message='Tasks containing shell code, should not be prefixed with python in function header')

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER):
            try:
                ast.parse(item.FuncBodyRaw, 'tempfile')
            except Exception:
                if item.IsPython:
                    res += self.finding(item.Origin, item.InFileLine)
        return res



from oelint_adv.cls_rule import Rule


class NoSpaceEmptyLineRule(Rule):
    def __init__(self):
        super().__init__(id='oelint.spaces.emptyline',
                         severity='warning',
                         message='Empty lines shall not contain spaces')

    def __getMatches(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file)
        for i in items:
            if i.Raw.strip('\n') and not i.Raw.strip():
                res.append(i)  # pragma: no cover
        return res

    def check(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            res += self.finding(i.Origin, i.InFileLine)  # pragma: no cover
        return res

    def fix(self, _file, stash):  # pragma: no cover
        res = []
        for i in self.__getMatches(_file, stash):
            i.RealRaw = '\n'
            i.Raw = '\n'
            res.append(_file)
        return res



from colorama import Fore
from colorama.ansi import AnsiCodes

_COLOR = False
_COLORS_BY_SEVERITY = {
    'info': Fore.GREEN,
    'warning': Fore.YELLOW,
    'error': Fore.RED,
}


def get_colorize() -> bool:
    """Returns weather or not the terminal output is to be colorized"""
    return _COLOR


def set_colorize(value: bool) -> None:
    """Globally turn colored terminal output on/off"""
    global _COLOR
    _COLOR = value


def get_color_by_severity(severity: str) -> AnsiCodes:
    """Get an ANSI color code given a severity

    Args:
        severity (str): either `neutral`, `info`, `warning` or `error`
    """
    return _COLORS_BY_SEVERITY.get(severity, '')



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.constants import CONSTANTS
from oelint_parser.helper_files import is_image
from oelint_parser.helper_files import is_packagegroup


class VarMandatoryExists(Rule):
    def __init__(self):
        super().__init__(id='oelint.var.mandatoryvar',
                         severity='error',
                         message='<FOO>',
                         onappend=False,
                         appendix=CONSTANTS.VariablesMandatory)

    IMAGE_EXCLUDES = [
        'CVE_PRODUCT',
        'HOMEPAGE',
        'SRC_URI',
    ]

    PACKAGEGRP_EXCLUDES = [
        'CVE_PRODUCT',
        'HOMEPAGE',
        'LICENSE',
        'SRC_URI',
    ]

    def check(self, _file, stash):
        res = []
        _is_pkg_group = is_packagegroup(stash, _file)
        _is_image = is_image(stash, _file)
        for var in CONSTANTS.VariablesMandatory:
            if _is_pkg_group and var in VarMandatoryExists.PACKAGEGRP_EXCLUDES:
                continue
            if _is_image and var in VarMandatoryExists.IMAGE_EXCLUDES:
                continue
            items = stash.GetItemsFor(
                filename=_file, classifier=Variable.CLASSIFIER, attribute=Variable.ATTR_VAR, attributeValue=var)
            if not any(items):
                res += self.finding(_file, 0,
                                    'Variable \'{a}\' should be set'.format(a=var), appendix=var)
        return res



from oelint_adv.cls_rule import Rule


class NewLineEOF(Rule):
    def __init__(self):
        super().__init__(id='oelint.newline.eof',
                         severity='warning',
                         message='File shall end on a newline')

    def __getMatches(self, _file, stash):
        res = {}
        items = stash.GetItemsFor(filename=_file)
        for _uniqname in {x.Origin for x in items}:
            res[_uniqname] = sorted(stash.GetItemsFor(
                filename=_uniqname, nolink=True), key=lambda x: x.InFileLine)
        return res

    def check(self, _file, stash):
        res = []
        for _, v in self.__getMatches(_file, stash).items():
            if not v[-1].Raw.endswith('\n'):
                res += self.finding(v[-1].Origin, v[-1].InFileLine)
        return res

    def fix(self, _file, stash):
        res = []
        for f, v in self.__getMatches(_file, stash).items():
            if not v[-1].RealRaw.endswith('\n'):
                v[-1].RealRaw += '\n'
                v[-1].Raw += '\n'
                res.append(f)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import expand_term


class VarTrailingSlash(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.notrailingslash',
                         severity='error',
                         message='\'{a}\' must not end with a \'/\'')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        _needles = ['S', 'B', 'T', 'D']
        for i in items:
            if i.VarName not in _needles:
                continue
            _expanded = expand_term(stash, _file, i.VarValueStripped)
            if _expanded.endswith('/'):
                res += self.finding(i.Origin, i.InFileLine,
                                    override_msg=self.Msg.format(a=i.VarName))
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarAppendOperation(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.appendop',
                         severity='error',
                         message='Use \'{a}\' instead of \'{b}\' as it overwrites \'{c}\'')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        _names = {x.VarName for x in items if x.VarName != 'inherit'}
        for name in _names:
            _weakest = [x for x in items if x.VarName == name and x.VarOp == ' ??= ']
            _weak = [x for x in items if x.VarName == name and x.VarOp == ' ?= ']
            _items = [x for x in items if x.VarName == name and x not in _weakest + _weak and x.VarOp in [' .= ', ' += ']]
            for i in _items:
                if any(_weakest):
                    res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format(
                        a='_append', b=i.VarOp, c=_weakest[0].Raw))
                elif any(x.Line > i.Line for x in _weak):
                    res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format(
                        a='_append', b=i.VarOp, c=_weak[0].Raw))
            _items = [x for x in items if x.VarName == name and x not in _weakest + _weak and x.VarOp in [' =. ', ' =+ ']]
            for i in _items:
                if any(_weakest):
                    res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format(
                        a='_prepend', b=i.VarOp, c=_weakest[0].Raw))
                elif any(x.Line > i.Line for x in _weak):
                    res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format(
                        a='_prepend', b=i.VarOp, c=_weak[0].Raw))
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.helper_files import get_scr_components
from oelint_parser.parser import INLINE_BLOCK


class VarSRCURIWildcard(Rule):
    def __init__(self):
        super().__init__(id='oelint.var.srcuriwildcard',
                         severity='error',
                         message='\'SRC_URI\' should not contain any wildcards')

    def check(self, _file, stash):
        res = []
        _items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')
        for item in _items:
            for f in [x.strip('\'') for x in item.get_items() if x]:
                if f == INLINE_BLOCK:
                    continue
                components = get_scr_components(f)
                if components['scheme'] == 'file':
                    if any([x for x in ['*'] if x in components['src']]):
                        res += self.finding(item.Origin, item.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.helper_files import expand_term
from oelint_parser.helper_files import get_valid_package_names


class VarPkgSpecific(Rule):
    def __init__(self):
        self.needles = ['RDEPENDS', 'RRECOMMENDS', 'RSUGGESTS', 'RCONFLICTS', 'RPROVIDES', 'RREPLACES',
                        'FILES', 'pkg_preinst', 'pkg_postinst', 'pkg_prerm', 'pkg_postrm', 'ALLOW_EMPTY']
        super().__init__(id='oelint.vars.pkgspecific',
                         severity='error',
                         message='Variable {VAR} is package-specific and therefore it should be {VAR}_${PN} or {VAR}:${PN}',
                         appendix=self.needles)

    def check(self, _file, stash):
        res = []

        if self.is_lone_append(stash, _file):
            return res

        _packages = list(get_valid_package_names(stash, _file))
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if i.VarName == 'inherit':
                continue
            if i.VarName in self.needles:
                _machine = []
                if i.GetMachineEntry():
                    _machine = [i.GetMachineEntry(), expand_term(
                        stash, _file, i.GetMachineEntry())]
                if not _machine or not any(x in _packages for x in _machine):
                    res += self.finding(i.Origin, i.InFileLine,
                                        override_msg=self.Msg.replace('{VAR}', i.VarName), appendix=i.VarName)
        return res



import re

from oelint_adv.cls_rule import Rule


class NoSpaceTrailingRule(Rule):
    def __init__(self):
        super().__init__(id='oelint.spaces.lineend',
                         severity='warning',
                         message='Line shall not end with a space')

    def __getMatches(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file)
        for i in items:
            _linecnt = 0
            for line in i.Raw.split('\n'):
                if line.endswith(' '):
                    res.append((i, i.Line + _linecnt))
                _linecnt += 1
        return res

    def check(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            res += self.finding(i[0].Origin, i[1])
        return res

    def fix(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            i.RealRaw = re.sub(r'\s{2,}\n', '\n', i[0].RealRaw)
            i.Raw = re.sub(r'\s{2,}\n', '\n', i[0].Raw)  # pragma: no cover
            res.append(_file)  # pragma: no cover
        return res  # pragma: no cover



from urllib.error import HTTPError
from urllib.error import URLError
from urllib.request import Request
from urllib.request import urlopen

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarHomepagePing(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.homepageping',
                         severity='warning',
                         message='\'HOMEPAGE\' isn\'t reachable')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='HOMEPAGE')
        for i in items:
            try:
                req = Request(i.VarValueStripped)  # noqa: S310 - we can take the risk of calling unexpected schemes here
                try:
                    urlopen(req, timeout=4)  # noqa: S310 - we can take the risk of calling unexpected schemes here
                except HTTPError as e:
                    if e.code == 404:  # pragma: no cover
                        res += self.finding(i.Origin,
                                            i.InFileLine)  # pragma: no cover
                except URLError:
                    res += self.finding(i.Origin, i.InFileLine)
            except ValueError:
                res += self.finding(i.Origin, i.InFileLine)
            except Exception:  # pragma: no cover
                pass  # pragma: no cover
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Function
from oelint_parser.constants import CONSTANTS


class TaskOrder(Rule):
    def __init__(self):
        super().__init__(id='oelint.task.order',
                         severity='warning',
                         message='<FOO>',
                         appendix=CONSTANTS.FunctionsOrder)

    def check(self, _file, stash):
        res = []
        _files = {item.Origin for item in stash.GetItemsFor(
            filename=_file, classifier=Function.CLASSIFIER)}

        for _single_file in _files:
            items = stash.GetItemsFor(
                filename=_single_file, classifier=Function.CLASSIFIER, nolink=True)
            for item in items:
                _func_before = sorted(
                    [x for x in items if x.Line < item.Line and x.FuncName in CONSTANTS.FunctionsOrder], key=lambda x: x.Line, reverse=True)
                if any(_func_before):
                    _func_before = _func_before[0]
                    if item.FuncName not in CONSTANTS.FunctionsOrder:
                        continue
                    if CONSTANTS.FunctionsOrder.index(item.FuncName) < CONSTANTS.FunctionsOrder.index(_func_before.FuncName):
                        res += self.finding(item.Origin, item.InFileLine,
                                            '\'{a}\' should be placed before \'{b}\''.format(
                                                a=item.FuncName, b=_func_before.FuncName), appendix=item.FuncName)
        return res
